#include <cmath>
#include <vector>
using std::vector;
#include <algorithm>
#include <set>
using std::set;
#include <map>
using std::map;
#include <queue>
using std::priority_queue;
#include <stack>
#include "analysis.H"
#include "ldock.H"
#include <cassert>
#include <iostream>
using std::cout;
using std::endl;
using std::cerr;
#include "transform.H"
#include <fstream>
using std::ofstream;
#include <iomanip>
using namespace std;
#include "roadmapConstruction.H"

int Path::startIndex = -1, Path::endIndex = -1, Path::numElementsInTree = 0;
const int Path::uninitializedDeviatedEdgeIndexValue = -2;
const float Path::clusterRadius = 5.3;


extern float distanceBetweenNodes(atomData *a1, atomData *a2);
float findRMSD(atomData *a1, atomData *a2);

extern receptorClass myReceptor;
extern queType *myQue;
extern int nodeEnergyListCount;
extern neighbourType nodeEnergyList[];
extern nodeData  *allNodes[];
extern int numNodes;
extern float currentTime;
extern int mycompare(const void *vfirst, const void *vsecond);
extern int MIN_REGIONS;
extern int globalNumAtoms;
extern atomData *origAtomData;
extern int linkTimeArray[];
extern int totalExtraNodes;


/* 
   finds the best path from start to end using bfs.
   returns the path (that needs to be processed) in pred, which contains 
     the predecessor node for each node.
   also returns true if best path found. o.w. returns false.
*/
 
bool BFS (int sI, int eI, nodeData ** allNodes,
	  map<int,int> &pred, float * pW)
{
  priority_queue<elementType, vector<elementType>, prioritizeNodes> nodesQueue;
  set<int> visitedNodes;
  bool     targetFound     = false;
  map<int, float>  nodesAddedToPQ; // to keep pred up to date (same node 
  // may be added twice to the pq, via different paths, and its pred.
  // should always be the node which causes the fastest path to that node)
  // nodesAddedToPQ[i] = t means i'th node added to p.q. at the earliest time
  // of t.

  elementType e(0.0, sI);

  nodesQueue.push(e);
  nodesAddedToPQ[sI] = 0.0;

  while ((!nodesQueue.empty()) && (!targetFound))
    {
      elementType n          = nodesQueue.top();
      nodesQueue.pop();
      float currentTime      = n.getTime();
      int   currentNodeIndex = n.getIndex();

      if (visitedNodes.find(currentNodeIndex) != visitedNodes.end())
	continue; // found this node among those visited, skip.

      assert(visitedNodes.find(currentNodeIndex) == visitedNodes.end());
      //      cout << "currentTime = " << currentTime << " nodeIndex = " << currentNodeIndex << endl;
      
      visitedNodes.insert(currentNodeIndex);
      
      vector<int> neighborIndices;
      assert(neighborIndices.size() == 0);
      
      findNeighbors (neighborIndices, currentNodeIndex, allNodes);
      
      for (unsigned int i = 0; i < neighborIndices.size(); i++)
	{
	  int otherNodeIndex = neighborIndices[i];
	  if (visitedNodes.find(otherNodeIndex) == visitedNodes.end())
	    {
	      // node not visited yet.
	      // pushing it to pq.
	      
	      elementType q (currentTime + 
			     fPathWeight(currentNodeIndex,
					 otherNodeIndex, 
					 allNodes),
			     otherNodeIndex);
	      if (q.getTime() < currentTime)
		{
		  cout << "cNI : " << currentNodeIndex << " oNI : " ;
		  cout << otherNodeIndex << " time : " << currentTime ;
		  cout << "q.getTime() = " << q.getTime() << endl;
		}
	      assert(q.getTime() >= currentTime);
	      
	      nodesQueue.push(q);
	      map<int,float>::iterator m = nodesAddedToPQ.find(q.getIndex());
	      
	      if (m != nodesAddedToPQ.end())
		{
		  // this node was previously added to pq.
		  // will conditionally update the pred map.
		  if ( m->second > q.getTime())
		    {
		      m->second = q.getTime();
		      pred[otherNodeIndex] = currentNodeIndex;
		    }
		}
	      else
		{
		  nodesAddedToPQ [q.getIndex()] = q.getTime();
		  
		  pred[otherNodeIndex] = currentNodeIndex;
		}
	      
	      //      cout << "inserted a node of index : " << q.getIndex();
	      //      cout << " at time : " << q.getTime()  << endl; 
	      // setting its predecessor node index to currentNodeIndex.


	      if (otherNodeIndex == eI)
		{
		  targetFound = true;
		  (*pW)       = q.getTime();
		  break;
		}
	    }
	}
    }
  return targetFound;
}

/* 
   traverses back from the endIndex to startIndex, and pushes
   each index to the vector, then forms a path out of this.
   pW is the path weight.
   pred[i] contains the node that precedes i.
*/
void trackPath (Path ** bestPath, map<int,int>& pred, 
		int startIndex, int endIndex, float pW, 
		nodeData ** allNodes)
{

  
  int cN = endIndex; // currentNode is endIndex;
  vector<int> v;
  set<int> debugS; // for debugging

  while (cN != startIndex)
    {
      //      cout << cN << " ";
      assert(debugS.find(cN) == debugS.end());
      debugS.insert(cN);
      v.push_back(cN);
      cN = pred [cN];
    }
  v.push_back(cN);
  
  (*bestPath) = new Path (pW, v, allNodes);

}
/* 
   finds the best path from node #i to node #j.
   returns the bestpath, which includes first and last indices.
   
   returns null if no path found.
*/
void newFindBestPath (Path ** bestPath, int startIndex, int endIndex,
		      int numberOfNodes, nodeData ** allNodes)
{
  if (allNodes[startIndex]->myTree != allNodes[endIndex]->myTree) 
    {
      (*bestPath) = NULL;
      return;
    }
  
  if (startIndex == endIndex) 
    {
      vector<int> v;
      v.push_back(startIndex);
      assert(v.size() == 1);
      (*bestPath) = new Path (0.0, v, allNodes);
      return;
    }
  
  map<int,int> pred; // contains the predecessor of each node in the best path.
  float pW; // is a path is found, pathweight returned in pW.

  if (BFS(startIndex, endIndex, allNodes,pred, &pW))
    {
      // target found
      trackPath(bestPath, pred, startIndex, endIndex, pW, allNodes);
    }
  else
    {
      cout << "no path found." << endl;
      (*bestPath) = NULL;
    }
}


/* 
   constructor for path. vector is used as a stack, so the elements should
   be retrieved from the end of stack.
   
*/
Path::Path(float _pathWeight, const vector<int>& _indices,
	   nodeData ** allNodes)
{
  int i;
  

  numIndices = _indices.size();
  indices    = new int[numIndices];
  for (i = 0;  i < numIndices; i++) 
    {
      indices[i]    = _indices[numIndices - i - 1];
    }

  pathWeight        = _pathWeight;
  children          = NULL; 
  numChildren       = 0;
  found             = false;
  deviatedEdgeIndex = uninitializedDeviatedEdgeIndexValue;
  
  updatePathWeight(allNodes);
  if (pathWeight != _pathWeight) 
    {
      cout << "input path weight = " << _pathWeight << " updated pathweight = " << pathWeight << endl;
      exit(0);
    }
}

void Path::updatePathWeight(nodeData ** allNodes) {
  float temp = 0.0;
  int i;

  for (i = 0; i < numIndices - 1; i++) {
    //    cout << indices[i] << " -> " << indices[i+1] << " ";
    temp += fPathWeight(indices[i], indices[i+1], allNodes);
    //    cout << temp << endl;
  }
  pathWeight = temp;
}

/* 
   returns forward path weight between allNodes[startIndex] and 
   allNodes[endIndex].
*/
float fPathWeight(int startIndex, int endIndex, nodeData ** allNodes) 
{
  
  float       retval  = -1.0;
  ptrList   * linkPtr = allNodes[startIndex]->myNode->firstLink;
  nodeClass * endNode = allNodes[endIndex]  ->myNode;
  
  while (linkPtr != NULL) {
    if (linkPtr->node == endNode) {
      retval = linkPtr->pathWeight;
      break;
    }
    linkPtr = linkPtr->nextPtr;
  }

  if (retval < 0) 
    {
      assert(retval == -1.0);
      cout << "forward path weight could not find path between " << startIndex << " and " << endIndex << endl;
      vector<int> neighborIndices;
      findNeighbors(neighborIndices, startIndex, allNodes);
      
      for (unsigned int i = 0; i < neighborIndices.size(); i++) 
	{
	  cerr << neighborIndices[i] << " ";
	}
      cerr << endl;
      
      findNeighbors(neighborIndices,  endIndex, allNodes);
      for (unsigned int i = 0; i < neighborIndices.size(); i++) 
	{
	  cerr << neighborIndices[i] << " ";
	}
      cerr << endl;
      

    //    retval = 0.1;
    assert(0);
    }

  return retval;  
}


/* 
   neighbors should not be allocated. Allocated here. 
   if numNeighbors > maxArraySize, exits.
*/

void findNeighbors (vector<int> &neighborIndices, int startNode, 
		    nodeData ** allNodes) 
{
  
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  while (linkPtr != NULL) 
    {
      neighborIndices.push_back(linkPtr->node->myConfig->index);
      linkPtr                     = linkPtr->nextPtr;
    }
}


void Path::print(nodeData ** allNodes)
{
  
  int i;

  cout << "***********************************************" << endl;

  cout << "path weight = " << pathWeight << endl;
  cout << "Indices = " ;
  for (i = 0; i < numIndices; i++) {
    cout << i << "- " << indices[i] << " , ";
  }
  cout << endl;

  cout << "Energies : " ;
  for  (i = 0; i < numIndices; i++) {
    cout << allNodes[indices[i]]->nodeEnergy << "  ";
  }
  cout << endl;

  cout << "edge weights : " << endl;
  for (i = 0; i < numIndices - 1; i++) {
    cout << fPathWeight(indices[i], indices[i+1], allNodes) << " ";
  }
  cout << endl;
  //compare visually nodes passed;
}

void Path::printEnergyProfile(nodeData ** allNodes, FILE * fout)
{
  for  (int i = 0; i < numIndices; i++) 
    {
      fprintf(fout, "%f ", allNodes[indices[i]]->nodeEnergy);
    }
  fprintf(fout, "\n");
}

/*
  finds a random node that is at least at a distance of d from 
  protein surface. returns its index.
  returns -1 if no such node is foiund after numIter iterations.
*/

int findStartNode(float d, nodeData ** allNodes, int numNodes )
{
  int       retval        = -1, numIter = 0, startNodeIndex, maxIter = 30;
  bool      found         = false;
  
  while ((numIter < maxIter) && (!found))
    {
      startNodeIndex = (int) (drand48() * numNodes);
      float surfDist = myReceptor.distance(allNodes[startNodeIndex]->cog);

      if (surfDist > d)
	{
	  found = true;
	  cout << "found start node. distance from surface : " << surfDist << endl;
	  retval = startNodeIndex;
	}
      numIter ++;
    }
  
  return retval;
}

/*
  finds the shortest path to the potential binding sites, starting from
  a node far from the protein surface. (distance > d).
  Prints them in a file with the given file name.
*/
void printEnergyProfiles2BindingSites(const vector<int> &
				      potBindingSiteIndices,   
				      const char * filename,   
				      nodeData ** allNodes,
				      int numNodes)
{
  int startNodeIndex, endNodeIndex;
  const float d = 10.0; // minDistanceToProteinSurface
  if ((startNodeIndex = findStartNode(d, allNodes, numNodes)) == -1)
    {
      cout << "couldn't find a start node. returning... " << endl;
      return;
    }

  Path * bestPath;
  FILE * f_EnergyProfile;
  f_EnergyProfile = fopen("energyProfile.txt", "w");
  cout << " check energyProfile.txt" << endl;

  for (int i = 0; i < potBindingSiteIndices.size(); i++)
    {
      endNodeIndex = potBindingSiteIndices[i];
      cout << " end node info: # : " << endNodeIndex << " energy = " << allNodes[endNodeIndex]->nodeEnergy << endl;
      cout << "distance of start node to end node : " ;
      cout << distanceBetweenNodes(allNodes[startNodeIndex]->myData, 
				   allNodes[endNodeIndex]->myData) << endl;
      newFindBestPath (&bestPath, startNodeIndex, endNodeIndex, 
		       numNodes, allNodes);

      if (bestPath != NULL) 
	{
	  bestPath->print(allNodes);
	  bestPath->printEnergyProfile(allNodes, f_EnergyProfile);
	}
      else 
	{
	  cout << "no path found from : " << startNodeIndex <<  " to " << endNodeIndex << endl;
	}
    }
  fclose(f_EnergyProfile);
}

/*
  
  if pairwise distance between two centers < pairwiseRMSDThreshold,
  or the distance of a funnel center to the protein > protSurfDistThreshold,
  we remove that funnel center and not add it to the list of funnels.
  
  the funnel centers are taken from sampledAroundList, and does not include
  true binding site. the first element in sampledarounslist is assumed to 
  be node 0, which is the true binding site. But in filtering, true binding 
  site is used to eliminate other pot. binding sites close to it.


*/

void filterIndices  (vector<int> & filteredIndices, int numSampledAround, 
		     neighbourType * sampledAroundList, nodeData ** allNodes,
		     float protSurfDistThreshold, float pairwiseRMSDThreshold)
{
  
  vector<bool> eliminated; // is true for i'th center if i'th center is eliminated

  eliminated.resize(numSampledAround);
  for (unsigned int i = 0; i <  eliminated.size(); i++)
    {
      eliminated[i] = false;
    }
  /*
  for (int i = 0; i < numSampledAround; i++)
    {
      if (sampledAroundList[i].index == 0)
	{
	  eliminated[i] = true;
	  break;
	}
    }
  */

  cout << "filtering funnel centers with a pairwise RMSD threshold of " <<  pairwiseRMSDThreshold << " and distance to protein surface of " << protSurfDistThreshold << endl;
  
  assert (sampledAroundList[0].index == 0);
  for (int i = 0; i < numSampledAround; i++)
    {
      for (int j = i+1; j < numSampledAround; j++)
	{
	  int node1 = sampledAroundList[i].index;
	  int node2 = sampledAroundList[j].index;

	  float rmsd  = findRMSD(allNodes[node1]->myData, 
				 allNodes[node2]->myData);
	  
	  if (rmsd < pairwiseRMSDThreshold)
	    {
	      if ((eliminated[i] == true) || (eliminated[j] == true))
		{
		  //dont do anything since one of them is eliminated already
		}
	      else
		{
		  eliminated[j]     = true;
		}
	    }
	}
    }

  for ( int i = 0; i < numSampledAround; i++)
    {
      if (eliminated[i])
	continue;
      int  node1 = sampledAroundList[i].index;
      float dist = myReceptor.distance(allNodes[node1]->cog);
      if (dist > protSurfDistThreshold)
	{
	  eliminated[i] = true;
	}
    }

  filteredIndices.resize(0);
  // note: index starts from 1!
  for (int i = 1; i < numSampledAround; i++)
    {
      if (eliminated[i] == false)
	{
	  filteredIndices.push_back(sampledAroundList[i].index);
	}
    }

  cout << "after filtering, " << filteredIndices.size() << " funnel centers remain." << endl;
}
/*
  returns indices of nodes closer than rmsdRadius to the original node.
  indices are 0-based.
*/
void findNodesClose2Node(vector<int> & v, int origIndex, int numNodes, 
			 nodeData ** allNodes, float rmsdRadius)
{
  v.resize(0);
  assert (origIndex < numNodes);
  assert (origIndex >= 0);

  for (int i = 0; i < numNodes; i++)
    {
      if (findRMSD(allNodes[i]->myData, 
		   allNodes[origIndex]->myData) < rmsdRadius)
	{
	  //   put node i in the vector to be printed for node # nodeEnergyList[j]]->index
	  v.push_back(i);
	}
    }
}

void findNeighborsAndEdgeWeights (vector<int>   &neighborIndices,
				  vector<float> &edgeWeights,
				  int startNode,
				  int numNodes, 
				  nodeData ** allNodes)
{
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  edgeWeights    .resize(0);

  while (linkPtr != NULL) 
    {
      int index = linkPtr->node->myConfig->index;
      neighborIndices.push_back (index);
      if (index > numNodes)
	{
	  cout  << "index = " << index << " numnodes = " << numNodes << endl;
	}
      assert         (index < numNodes);
      edgeWeights    .push_back (linkPtr->pathWeight);
      linkPtr       = linkPtr->nextPtr;
    }
}


unsigned int findNumNeighbors (int i, int numNodes, nodeData ** allNodes)
{
  ptrList * linkPtr;
  unsigned int retval = 0;
  linkPtr  = allNodes[i]->myNode->firstLink;

  assert (i < numNodes);
  while (linkPtr != NULL) 
    {
      int index = linkPtr->node->myConfig->index;
      retval++;
      if (index > numNodes)
	{
	  cout  << "index = " << index << " numnodes = " << numNodes << endl;
	}
      assert         (index < numNodes);
      linkPtr       = linkPtr->nextPtr;
    }

  return retval;
}


/*
  seed file contains the indices of nodes of interest.
  this function will go thru each node in the roadmap and 
  write the rmsd distance to each of the seeds to a file,
  columnwise.
*/
void analyzeRMSDDistanceFromSeeds(const char * seedFilename, int numNodes, nodeData ** allNodes)
{
  FILE * fin = fopen (seedFilename, "r");
  vector<int> seedIndices;

  cout << "reading " << seedFilename << endl;

  int    index;

  if (fin == NULL)
    {
      cout << "couldn't open " << seedFilename << endl;
      exit(0);
    }

  while (!feof(fin))
    {
      fscanf  (fin, "%d\n", &index);
      cout << "seed index = " << index << endl;
      seedIndices.push_back(index);
    }
  
  fclose (fin);

  FILE * fout = fopen ("distancesFromSeeds.txt", "w");
  cout << "check distancesFromSeeds.txt" << endl;
  
  for (int i = 0; i < numNodes; i++)
    {
      for (unsigned int j = 0; j < seedIndices.size(); j++)
	{
	  fprintf( fout, "%f ", findRMSD(allNodes[i]->myData, 
					 allNodes[seedIndices[j]]->myData) );
	}
      fprintf(fout, "\n");
    }
  
  fclose (fout);
}

/*
  reads file and prints the rmsd distance of the nodes from the tbs to the file
  distancesFromTBS.txt.
*/

void analyzeHighStatDistNodes (const char * indexFilename, 
			       nodeData ** allNodes,
			       atomData * origAtomData)
{
  FILE * fin = fopen (indexFilename, "r");
  
  cout << "reading " << indexFilename << endl;
  int    index;

  if (fin == NULL)
    {
      cout << "couldn't open " << indexFilename << endl;
      exit(0);
    }

  FILE * fout = fopen ("distancesFromTBS.txt", "w");
  cout << "check distancesFromTBS.txt" << endl;

  while (!feof(fin))
    {
      fscanf(fin, "%d\n", &index);
      cout << "index = " << index << endl;
      fprintf (fout, "%f\n", findRMSD(origAtomData, allNodes[index]->myData) );
      //      fscanf  (fin,  "%d\n", &index);
    }
  
  fclose (fin);
  fclose (fout);
}

///
void analyzePairwiseAtomDistances(int globalNumAtoms, const atomData *adata)
{
  ofstream fout ("connections.txt");
  const float maxCovalentBondLength = 1.7;

  fout.width(4);
  
  cerr << "check connections.txt for the ligand connection info. " << endl;
  for (int i = 0; i < globalNumAtoms; i++)
    {
      vector<int> closeNodes(0);
      for (int j = 0; j < globalNumAtoms; j++)
	{
	  if (j == i) continue;
	  float dist = distanceBetweenVecs(adata[i].xyz, adata[j].xyz);

	  if ((i == 18) && ( j == 26))
	    {
	      cout << "distance between " << i << " and " << j << " is : " << dist << endl;
	      cout << adata[j].xyz[0] << " " << adata[j].xyz[1] << " " << adata[j].xyz[2] << endl;
	    }

	  if (dist < maxCovalentBondLength)
	    {
	      //cout << "distance between " << i << " and " << j << " is : " << dist << endl;
	      closeNodes.push_back(j);
	    }
	}
      if (closeNodes.size() != 0)
	{
	  fout << "CONECT " << setw(4) << i+1;
	  for (vector<int>::iterator iter = closeNodes.begin(); iter != closeNodes.end(); iter++)
	    {
	      fout << " " << setw(4) << (*iter) + 1 ;
	    }
	  fout << endl;
	}
    }
}



void initNodeColor(int thecolor) {

  int i;

  
  for (i = 0; i < numNodes; i++) {
    allNodes[i]->myNode->nodeColor = (char)(thecolor);
    allNodes[i]->myNode->nodeTime = -1;
  }

}


void initQue(float t, nodeClass *fromN, nodeClass *toN) {

  myQue = new queType;

  myQue->timeStamp = t;
  myQue->fromNode = fromN;
  myQue->toNode = toN;
  myQue->nextQEntry = NULL;

}

void addToQue(float t, nodeClass *fromN, nodeClass *toN) {

  queType *tempQuePtr, *prevQuePtr, *newQuePtr;
  int done;
 
  tempQuePtr = myQue;
  prevQuePtr = NULL;

  done = 0;
  while ((tempQuePtr != NULL) && (done == 0)) {
    if (tempQuePtr->timeStamp <= t) {
      prevQuePtr = tempQuePtr;
      tempQuePtr = tempQuePtr->nextQEntry;
    }
    else {
      done = 1;
    }
  }

  newQuePtr = new queType;

  newQuePtr->timeStamp = t;
  newQuePtr->fromNode = fromN;
  newQuePtr->toNode = toN;

  newQuePtr->nextQEntry = tempQuePtr;
  prevQuePtr->nextQEntry = newQuePtr;
}

void runBFS(nodeClass *targetnode) {

  queType *tempQuePtr;
  int done;
  int foundtarget = 0;
  //int iterations = 0;

  while ((myQue != NULL) && (foundtarget == 0)) {

    currentTime = myQue->timeStamp;
    tempQuePtr = myQue;
    done = 0;
    while ((tempQuePtr != NULL) && (done == 0)) {
      if (tempQuePtr->timeStamp == currentTime) {
	if (tempQuePtr->toNode->nodeColor == (char)(0)) {
	  tempQuePtr->toNode->nodeColor = (char)(1); 
	}
	else {
	  tempQuePtr->toNode->nodeColor = (char)(2); 
	}

	tempQuePtr = tempQuePtr->nextQEntry;
      }
      else {
	done = 1;
      }
    }


    done = 0;
    while ((myQue != NULL) && (done == 0)) {
      if (myQue->timeStamp == currentTime) {
	if (myQue->toNode->nodeColor == (char)(1)) {
	  myQue->toNode->setPath(myQue->fromNode, currentTime);
	  //	  cout << iterations << " " << currentTime << " ";
	  //	  iterations++;
	  if (myQue->toNode == targetnode) {
	    foundtarget = 1;
	  }
	}
	tempQuePtr = myQue;
	myQue = myQue->nextQEntry;
	delete tempQuePtr;
      }
      else {
	done = 1;
      }
    }
  }

  if (myQue != NULL) {
    while (myQue != NULL) {
      tempQuePtr = myQue;
      myQue = myQue->nextQEntry;
      delete tempQuePtr;
    }
  }    

}




/*
  prepares node energy list, which contains an energy ordered list of nodes.
  the catalytic site is included in this computation.
*/
void findMinEnergyLigs() {

  int i;
  float minEnergy = FLT_MAX;
  float maxEnergy = -FLT_MAX;
  float avgEnergy = 0.0;

  // careful: this is index 1!

  for (i = 1; i < numNodes; i++) 
    {
      if (allNodes[i]->nodeEnergy < minEnergy) {
	minEnergy = allNodes[i]->nodeEnergy;
      }
      if (allNodes[i]->nodeEnergy > maxEnergy) {
	maxEnergy = allNodes[i]->nodeEnergy;
      }
      avgEnergy += allNodes[i]->nodeEnergy;
    }

  avgEnergy = avgEnergy/((float)(numNodes));
  
  cout << "Min Node Energy = " << minEnergy << endl;
  cout << "Max Node Energy = " << maxEnergy << endl;
  cout << "Avg Node Energy = " << avgEnergy << endl;
    
  nodeEnergyListCount = 0;


  //careful: this is index 1!!
  for (i = 1; i < numNodes; i++) {
    if (allNodes[i]->nodeEnergy < avgEnergy/5.0) {
      //    if (allNodes[i]->nodeEnergy <= maxEnergy) {
      nodeEnergyList[nodeEnergyListCount].dist = allNodes[i]->nodeEnergy;
      nodeEnergyList[nodeEnergyListCount].index = i;
      nodeEnergyListCount++;
    }
  }
  
  nodeEnergyList[nodeEnergyListCount].dist  = allNodes[0]->nodeEnergy;
  nodeEnergyList[nodeEnergyListCount].index = 0;
  nodeEnergyListCount++;
  
  //  ADDED NODE # 0 TO NODE ENERGY LIST.;
  
  cout << "Num Nodes with Energy < " << avgEnergy/5.0 << " is = " << nodeEnergyListCount << endl;

  qsort(nodeEnergyList, nodeEnergyListCount, sizeof(neighbourType), 
	mycompare);      
  
  /*
    this part prints the sorted energies, indices.
  */
  
  int numNodes2Print = (nodeEnergyListCount < MIN_REGIONS) ? nodeEnergyListCount : MIN_REGIONS;

  if (numNodes2Print > 0)
    cout << "sorted node energies are : " << endl;
  for (int i = 0; i < numNodes2Print; i++)
    {
      cout << "i = " << i << " index = " << nodeEnergyList[i].index << " energy = " << allNodes[nodeEnergyList[i].index]->nodeEnergy << endl;
    }

  cout << "0th nodes energy = " << allNodes[0]->nodeEnergy << endl;
}

///
float findRMSD(atomData *a1, atomData *a2) {

  int i;
  float retval = 0.0;

  for (i = 0; i < globalNumAtoms; i++) {
    retval = retval + (a1[i].xyz[0] - a2[i].xyz[0])*(a1[i].xyz[0] - a2[i].xyz[0]) + 
             (a1[i].xyz[1] - a2[i].xyz[1])*(a1[i].xyz[1] - a2[i].xyz[1]) + 
             (a1[i].xyz[2] - a2[i].xyz[2])*(a1[i].xyz[2] - a2[i].xyz[2]);
  }

  retval = sqrt(retval/((float)(globalNumAtoms)));

  return retval;
}


/*
  seed file contains the indices of nodes of interest.
  this function will go thru each node in the roadmap and 
  write the rmsd distance to each of the seeds to a file,
  columnwise.
*/
void analyzeRMSDDistanceFromSeeds(const char * seedFilename)
{
  FILE * fin = fopen (seedFilename, "r");
  vector<int> seedIndices;

  cout << "reading " << seedFilename << endl;

  int    index;

  if (fin == NULL)
    {
      cout << "couldn't open " << seedFilename << endl;
      exit(0);
    }

  while (!feof(fin))
    {
      fscanf  (fin, "%d\n", &index);
      cout << "seed index = " << index << endl;
      seedIndices.push_back(index);
    }
  
  fclose (fin);

  FILE * fout = fopen ("distancesFromSeeds.txt", "w");
  cout << "check distancesFromSeeds.txt" << endl;
  
  for (int i = 0; i < numNodes; i++)
    {
      for (unsigned int j = 0; j < seedIndices.size(); j++)
	{
	  fprintf( fout, "%f ", findRMSD(allNodes[i]->myData, 
					 allNodes[seedIndices[j]]->myData) );
	}
      fprintf(fout, "\n");
    }
  
  fclose (fout);
}

/*
  reads file and prints the rmsd distance of the nodes from the tbs to the file
  distancesFromTBS.txt.
*/

void analyzeHighStatDistNodes (const char * indexFilename)
{
  FILE * fin = fopen (indexFilename, "r");
  
  cout << "reading " << indexFilename << endl;
  int    index;

  if (fin == NULL)
    {
      cout << "couldn't open " << indexFilename << endl;
      exit(0);
    }

  FILE * fout = fopen ("distancesFromTBS.txt", "w");
  cout << "check distancesFromTBS.txt" << endl;

  while (!feof(fin))
    {
      fscanf(fin, "%d\n", &index);
      cout << "index = " << index << endl;
      fprintf (fout, "%f\n", findRMSD(origAtomData, allNodes[index]->myData) );
      //      fscanf  (fin,  "%d\n", &index);
    }
  
  fclose (fin);
  fclose (fout);
}

void checkPathEnergies(atomData **myPathConfigs, int plen, int sIndex, int eIndex, int *tcreated, int *tdestroyed) {

  int retval = 0;
  float myForwardLength = 0.00001;
  float myBackwardLength = 0.00001;
  int currentForwardIndex,j;// currentBackwardIndex, j;
  float temp1, temp2;
  float prevEnergy, currentEnergy, nextEnergy;
  nodeClass *newNode;

  if (plen == 1) {
    retval = makeLinkBetweenNodes(sIndex, eIndex, myForwardLength, myBackwardLength);
    if (retval == 1) {
      *tcreated = *tcreated + 1;
    }
    else if (retval == -1) {
      *tdestroyed = *tdestroyed + 1;
    }
  }

  else if (plen > 1) {

    prevEnergy = allNodes[sIndex]->nodeEnergy;
    currentEnergy = myReceptor.getEnergy(myPathConfigs[0]);

    currentForwardIndex = -1;
    
    if (currentEnergy < 0.0 /*MAX_ENERGY*/) {
      for (j = 0; j < plen - 1; j++) {
	currentForwardIndex = j;
	nextEnergy = myReceptor.getEnergy(myPathConfigs[j+1]);	  
	if (nextEnergy < 0.0 /*MAX_ENERGY*/) {
	  temp1 = exp((currentEnergy - nextEnergy)*INV_BOLTZMAN_K);
	  temp2 = exp((currentEnergy - prevEnergy)*INV_BOLTZMAN_K);
	  if (temp1 == 0.0) {
	    myForwardLength += (MAX_PATH_LENGTH)*ENERGY_WEIGHT + LENGTH_WEIGHT;
	  }
	  else {
	    myForwardLength += (-log(temp1/(temp1+temp2)))*ENERGY_WEIGHT + LENGTH_WEIGHT;
	  }
	  if (temp2 == 0.0) {
	    myBackwardLength += (MAX_PATH_LENGTH)*ENERGY_WEIGHT + LENGTH_WEIGHT;
	  }
	  else {
	    myBackwardLength += (-log(temp2/(temp1+temp2)))*ENERGY_WEIGHT + LENGTH_WEIGHT;
	  }
	  prevEnergy = currentEnergy;
	  currentEnergy = nextEnergy;
	  
	}
	else {
	  break;
	}
      }
    }
    if ((myForwardLength < MAX_PATH_LENGTH) || (myBackwardLength < MAX_PATH_LENGTH)) {
      if (currentForwardIndex == plen - 2) {
	retval = makeLinkBetweenNodes(sIndex, eIndex, myForwardLength, myBackwardLength);
	if (retval == 1) {
	  *tcreated = *tcreated + 1;
	}
	else if (retval == -1) {
	  *tdestroyed = *tdestroyed + 1;
	}
      }
      else if (currentForwardIndex != -1) {
	allNodes[numNodes] = copyConfig(myPathConfigs[currentForwardIndex]);
	newNode = new nodeClass;
	newNode->myConfig = allNodes[numNodes];
	allNodes[numNodes]->myNode = newNode;
	allNodes[numNodes]->nodeEnergy = currentEnergy;
	linkTimeArray[numNodes] = sIndex;
	numNodes++;
	totalExtraNodes++;

	retval = makeLinkBetweenNodes(sIndex, numNodes-1, myForwardLength, myBackwardLength);
	if (retval == 1) {
	  *tcreated = *tcreated + 1;
	}
	else if (retval == -1) {
	  *tdestroyed = *tdestroyed + 1;
	}
      }
    }
	  

  }
}




    




