/**@name funnelFind.C */
//@{

///
void findPotBindingSites(vector<int> & potBindingSiteIndices,
			 nodeData ** allNodes,
			 int   maxNumFunnelCenters,
			 float pairwiseRMSDThreshold,
			 float protSurfDistThreshold)
			 
{

  vector<int>  lowEnergySiteIndices;
  
  findDistinctLowEnergyConfigs(lowEnergySiteIndices,
			       allNodes,
			       maxNumFunnelCenters,
			       pairwiseRMSDThreshold);
  

  filterIndices  (potBindingSiteIndices,
		  lowEnergySiteIndices,
		  allNodes, 
		  protSurfDistThreshold);
  //according to their distance from 
  // the protein surface.
}

/** 
    this is the wrapper which calls other functions to find and print the 
    potential binding sites. 

    The indices of these files are returned
    in the potBindingSiteIndices vector, and they are also 
    written to the funnelCentersFile. 

    
    The actual number of funnel centers that is returned may be less
    than maxNumFunnelCenters. (see below)

    protSurfDistThreshold :  the max distance of funnel
    centers (potential binding sites) from the protein surface.

    pairwiseRMSDThreshold :  the min. pairwise distance
    between computed funnel centers.

    ps. the number of funnel centers computed and printed to the output file
    may be less than maxNumFunnelCenters since the functions find maxNumFunnelCenters
    low energy distinct sites, but then these are filtered according to
    their distance from the protein and the resulting number of funnel centers 
    are returned. Minor implementation issue to be dealt with in the future
    versions, by also considering the distan
    there may not be enough number of funnel centers that is found. Try
    increasing the input numFunnelCenters to (possibly)
    find more potential binding sites.

*/
void findAndPrintPotBindingSites(vector<int> & potBindingSiteIndices,
				 const char * funnelCentersFile, 
				 int          maxNumFunnelCenters,
				 float protSurfDistThreshold = 5.0,
				 float pairwiseRMSDThreshold = 10.0)
{
  assert (numFunnelCenters > 0);
  
  findPotBindingSites();
  
  printFunnelCenterInfo(potBindingSiteIndices,
			allNodes);
}


/**
   finds configs that are of low energy, that are distant from 
   each other. Some of these configs may actually be far away from
   the protein surface.
   
   These nodes are kept in sampledAroundList array.

   notes:

   MIN_REGIONS (the number of distinct low energy configs sought) should be non-0.
   At the end of this function, the nodes are ordered in energy.

   parameters:
   
   the implicit input parameters are INIT_NODES (which is the number of nodes to be sampled), MIN_REGIONS (which is the number of funnels to be explored).
   the implicit output parameter is sampledAroundList[i] (which holds the indices of funnel centers, as well as their energies)
*/
void findDistinctLowEnergyConfigs(float minRMSDBetweenFunnelCenters)
		   
{
  
  vector<int> n; 
  // n[i] holds the index of the first node that is sampled
  // around potBindingSiteIndices[i]. 
  // all nodes sampled around a potential binding site are consecutive 
  // in allNodes array, and end at n[i+1]-1.
  
  n.resize     (MIN_REGIONS);
  // sampling the nodes. 
  createRandomNodes  (INIT_NODES - numNodes);

  assert       (numNodes == INIT_NODES); 
  assert       (MIN_REGIONS != 0);
  
  sampleAroundDistinctLowEnergyConfigs(allNodes, minRMSDBetweenFunnelCenters,
				       sampledAroundList);
	    
  //      this qsort sorts on energy;
  qsort(&sampledAroundList, numSampledAround, sizeof(neighbourType), 
	mycompare);  
  
  cout << "lowest energy nodes are kept in an array. This array has the following entries for this protein:" << endl;
  cout << "contents: format: <index> <energy> " << endl;
  for (int i = 0; i < numSampledAround; i++)
    {
      cout << sampledAroundList[i].index << " " << sampledAroundList[i].dist << endl;
    }
  cout << "Sampling complete.  NumNodes = " << numNodes << endl;
      
  setIndices(numNodes); // necessary so as to have the index field of every node to be properly set.
}


/**
   finds funnel centers that are far apart from each other and samples
   around each SAMPLES_PER_REGION nodes.

   one funnel center is the node 0 in allNodes array, by default.

   parameters are: (some of them are global variables and are thus implicit)

   MIN_REGIONS funnels are to be explored.
   In each funnel, SAMPLES_PER_REGION nodes to be sampled.
   The centers of sampling are separated by at least minRMSDBetweenFunnelCenters. This is the min RMSD distance between ligands at the respective funnel centers. 
   sampledAroundList array, which contains funnel center node indices, as well
as their energies, is (implicitly) returned. At the end of iterative minimization around each potential funnel center, the sampleAroundListArray does not necessarily have the ordered list of energies. 

  notes:
  nodes should already be sampled. This function only finds out of the sampled nodes the ones that have the lowest energy, and refines around them.

*/

void sampleAroundDistinctLowEnergyConfigs(float minRMSDBetweenFunnelCenters)
{
  int minIndex, tooclose,j,i;
  

  findMinEnergyLigs();
  
  for (i = 0; ((numSampledAround < MIN_REGIONS) && 
	       (i < nodeEnergyListCount)); i++) 
    {

      tooclose = 0;

      // determining whether the current potential funnel center
      // is very close to those considered so far.
      for (j = 0; j < numSampledAround; j++) {
	if (distanceBetweenVecs(allNodes[sampledAroundList[j].index]->cog, 
				allNodes[nodeEnergyList[i].index]->cog) 
	    < minRMSDBetweenFunnelCenters)
	  {
	    tooclose = 1;
	    break;
	  }
      }
    
      if (tooclose == 0) 
	{
	  // will sample around this potential funnel center.
	  sampledAroundList[numSampledAround].linkedBit = numNodes;
	  minIndex                                      = nodeEnergyList[i].index;   
		  
	  for (j = 0; j < SAMPLES_PER_REGION/10; j++) 
	    {
	      minIndex = sampleAroundNode(minIndex, 10);
	    }
	  sampledAroundList[numSampledAround].dist  = allNodes[minIndex]->nodeEnergy;
	  sampledAroundList[numSampledAround].index = minIndex;
	  numSampledAround++;
	}
    }
}


/**
  prepares node energy list, which contains an energy ordered list of nodes.
  the catalytic site is included in this computation.
*/
void findMinEnergyLigs() {

  int i;
  float minEnergy = FLT_MAX;
  float maxEnergy = -FLT_MAX;
  float avgEnergy = 0.0;

  // careful: this is index 1!

  for (i = 1; i < numNodes; i++) 
    {
      if (allNodes[i]->nodeEnergy < minEnergy) {
	minEnergy = allNodes[i]->nodeEnergy;
      }
      if (allNodes[i]->nodeEnergy > maxEnergy) {
	maxEnergy = allNodes[i]->nodeEnergy;
      }
      avgEnergy += allNodes[i]->nodeEnergy;
    }

  avgEnergy = avgEnergy/((float)(numNodes));
  
  cout << "not including the catalytic site," << endl;
  cout << "Min Node Energy = " << minEnergy << endl;
  cout << "Max Node Energy = " << maxEnergy << endl;
  cout << "Avg Node Energy = " << avgEnergy << endl;
    
  nodeEnergyListCount = 0;

  //careful: this is index 1!!
  for (i = 1; i < numNodes; i++) {
    if (allNodes[i]->nodeEnergy < avgEnergy/5.0) {
      //    if (allNodes[i]->nodeEnergy <= maxEnergy) {
      nodeEnergyList[nodeEnergyListCount].dist  = allNodes[i]->nodeEnergy;
      nodeEnergyList[nodeEnergyListCount].index = i;
      nodeEnergyListCount++;
    }
  }
  
  nodeEnergyList[nodeEnergyListCount].dist  = allNodes[0]->nodeEnergy;
  nodeEnergyList[nodeEnergyListCount].index = 0;
  nodeEnergyListCount++;
  
  //  ADDED NODE # 0 TO NODE ENERGY LIST.;
  
  cout << "Num Nodes with Energy < " << avgEnergy/5.0 << " is = " << nodeEnergyListCount << endl;

  qsort(nodeEnergyList, nodeEnergyListCount, sizeof(neighbourType), 
	mycompare);      
  
  /*
    this part prints the sorted energies, indices.
  */
  
  int numNodes2Print = (nodeEnergyListCount < MIN_REGIONS) ? nodeEnergyListCount : MIN_REGIONS;

  if (numNodes2Print > 0)
    cout << "sorted node energies are : " << endl;
  for (int i = 0; i < numNodes2Print; i++)
    {
      cout << "i = " << i << " index = " << nodeEnergyList[i].index << " energy = " << allNodes[nodeEnergyList[i].index]->nodeEnergy << endl;
    }

  cout << "0th nodes energy = " << allNodes[0]->nodeEnergy << endl;
}

//@}
