function escapeTimes = computeEscapeTimesWrap(numRegions)
%
%
%input files : close2PotBindingSite#%d.txt, %d: 0..numRegions-1.
%              (which contains the indices in 0-based)
%              transitionProbs.txt
%
%
%output files: escapeTimes.txt, allEscapeTimes.txt.
%
%
%
%what it does: prepares the necessary input data structures, calls
%computeEscapeTimes, writes the results to a file.
%prints both region centers escape times and all nodes escape times
%to two different files.
%if there are > 1 cc, returns after writing an empty output file.
%
%
%assumptions : region centers are already far from each other and
%close2 protein surface, and that region centers are 
%nodes 1..numRegions
%
%


regionCentersEscapeTimeOutputFilename   = 'escapeTimes.txt';
allEscapeTimesOutputFilename            = 'allEscapeTimes.txt';
inputFilePrefix  = 'close2PotBindingSite#';
inputFileSuffix  = '.txt';

load transitionProbs.txt

biggestCCNodes       = findBiggestCC(transitionProbs);

%fprintf(1, 'num nodes in biggest cc is %d\n', length(biggestCCNodes));

tmp1     = max(transitionProbs(:,1)) + 1;
tmp2     = max(transitionProbs(:,2)) + 1;
numNodes = max(tmp1, tmp2);

if (numNodes ~= length(biggestCCNodes))

  fprintf(1, 'writing an empty %s, and then exiting.\n', regionCentersOutputFilename);
  escapeTimes = [];
  writeEscapeTimes(regionCentersOutputFilename, escapeTimes, 1);

else

  [numConnections,n]          = size(transitionProbs);
  fprintf(1, 'num connections = %d\n', numConnections);

  Tsparse  = prepareTsparse (transitionProbs);

  regionCenters = [1:numRegions];
  regions       = cell(numRegions, 1);

    for i = 1:numRegions
      filename = sprintf('%s%d%s', inputFilePrefix,  i-1, inputFileSuffix);
      regions{i} = load (filename) + 1; %regions are 0-based.
      fprintf(1, 'read %s\n', filename);
      if (length(find(regions{i} == 1)) == 1)
	debug_regionsAreZeroOrdered = 1;
      end
    end
  
  
%  technique = 3; % for cgs
  technique = 1; % for direct
  [expectedTime,flags]  =  computeEscapeTimes(Tsparse, regions, technique);%,1);
  flags

  if ((technique == 3) & (length(find(flags ~= 0)) > 0))
    fprintf(1, 'at least one region escape time comp. did not converge.\n');
    fprintf(1, 'writing an empty %s, and then exiting.\n', regionCentersOutputFilename);
    
    escapeTimes = [];
    writeEscapeTimes(regionCentersOutputFilename, escapeTimes);%, 1);

    return;
  else
    
    fprintf(1, 'expected time to escape from each funnel center is as follows:\n');
    escapeTimes = zeros(1, length(regionCenters));
    
    for i = 1:length(regionCenters)
      regionCenterIndex = find(regions{i} == regionCenters(i));
      escapeTimes(i)    = expectedTime{i}(regionCenterIndex);
    end
    
    regionCenters
    escapeTimes

    writeRegionCenterEscapeTimes(regionCentersEscapeTimeOutputFilename, escapeTimes, 0);
    writeAllEscapeTimes(allEscapeTimesOutputFilename, expectedTime, ...
			regions, numNodes);
  end
end %if  (numNodes ~= length(biggestCCNodes)) in else si.

%---------------------------------

function writeRegionCenterEscapeTimes(filename, escapeTimes, emptyFlag)
%writes escape times (of the region centers) to a file. writes empty if escape times are
%empty and emptyFlag is set.

if ((emptyFlag) & (length(escapeTimes) ~= 0))
  fprintf(1, ['potential error! escape times is not empty but empty' ...
	      ' flag is set\n']);
  
  keyboard;
end

if ((~emptyFlag) & (length(escapeTimes) == 0))
  fprintf(1, ['potential error! escape times is empty but empty' ...
	      ' flag is not set\n']);
  
  keyboard;
end

fid = fopen (filename, 'w');


if (emptyFlag)
  fprintf     (fid, '%escape times is empty\n');
else
  fprintf     (fid, '%escape times: \n');
  fprintf     (fid, '%e\n', escapeTimes');
end


fclose(fid);

%---------------------------------

%the format for the all escape times file is:
%numNodes
%escapeTime_1
%escapeTime_2
%escapeTime_3
%...
%escapeTimes_n.

function writeAllEscapeTimes (allEscapeTimesOutputFilename, expectedTime, ...
			      regions, numNodes)

%assuming that region indices are sorted.
%assert_regionIndicesAreSorted(regions);

fid = fopen(allEscapeTimesOutputFilename, 'w');

fprintf(fid, '%d\n', numNodes);

for i = 1:numNodes
  %find which region node i is in, (say region r), what entry it is
  %in that region (say k).
  %find its corresponding escape time expectedTime{r}(k).
  %print it.
  [regionIndex,relativeIndex] = findRegionAndRelativeIndexOfNode(i, regions);

  %fprintf(1, 'found relative and region indices. check now.\n');
  %fprintf(1, 'regionindex = %d relativeindex = %d\n', regionIndex, ...
%	  relativeIndex);
%  if (regionIndex ~= -1)
%    fprintf(1, 'node index = %d check %d\n', i, ...
%	    regions{regionIndex}(relativeIndex));
%  else
%    fprintf(1, 'could not find node %d in list of nodes\n', i);
%  end
%  keyboard;

  if (relativeIndex ~= -1)
    fprintf(fid, '%f\n', expectedTime{regionIndex}(relativeIndex));
  else
    fprintf(fid, '0\n');
  end
end

fclose(fid);


%---------------------------------

%later write accelerated version of this, which would take
%advantage of the fact that the indices in the regions are sorted
%in searching for the index of the node.

%returns -1 if node is not in any of the regions.
function [regionIndex,relativeIndex] = findRegionAndRelativeIndexOfNode(index, regions)

regionIndex   = -1;
relativeIndex = -1;
for i = 1:length(regions)
  relativeIndex      = find(regions{i} == index);
  if (length(relativeIndex) ~= 0)
    regionIndex = i;
    return;
  end
end
relativeIndex = -1; %so as not to return empty for relativeIndex
