function [et, flag] = computeOneEscTime(T, r, solver, maxit, prof)
%  
% Compute the escape time from a given region
%  
%    T:      transition probability matrix for the (entire) roadmap
%    r:      indices of nodes in the region (a potential binding site)
%    solver: 1 for direct, 2 for naive, 3 for CG
%    maxit:  maxinum number of iterations for iterative solvers
%    prof:   flag for profiling  
%
%flag is:
%0 if cgs is used, and converges to the desired tolerance within maxit iterations.
%-1 if another technique used.
%1..4 if cgs used and did not converge for some reason.

T = T(r,r);                             % select relevant rows and columns
flag     = -1;
numNodes = size(T,1);
b = ones(numNodes,1);

% solve the equation   et = T * et + b
I = speye(numNodes); %MSA: this was n.
T = I-T;

if (solver==1)				% direct method
  if (prof) profile on; end;
  et = direct(T,b);
else					% iterative methods
  et0 = ones(numNodes,1);
  
  % Don't use the sparse representation for the vectors. It slows down
  % the solver.
  et0 = full(et0); %MSA
  b   = full(b);

  if (prof) profile on; end; 
  
  if  (solver==2)  			% jacobi 
    et = jacobi(T, b, [], maxit, et0);
  elseif (solver==3)			% conjugate gradient
    [et,flag] = cgs(T, b, [], maxit, [],[], et0) ;
  end

end

if (prof) 
  profile report;
  profile off;
end;
